=== modified file 'bzrlib/smart/protocol.py'
--- bzrlib/smart/protocol.py	2007-11-09 19:21:56 +0000
+++ bzrlib/smart/protocol.py	2007-11-09 20:49:54 +0000
@@ -207,6 +207,8 @@
         self._write_func(_encode_tuple(response.args))
         if response.body is not None:
             assert isinstance(response.body, str), 'body must be a str'
+            assert response.body_stream is None, (
+                'body_stream and body cannot both be set')
             bytes = self._encode_bulk_data(response.body)
             self._write_func(bytes)
         elif response.body_stream is not None:
@@ -214,6 +216,7 @@
 
 
 def _send_stream(stream, write_func):
+    write_func('chunked\n')
     _send_chunks(stream, write_func)
     write_func('END\n')
 
@@ -266,7 +269,7 @@
 
     def __init__(self):
         _StatefulDecoder.__init__(self)
-        self.state_accept = self._state_accept_expecting_length
+        self.state_accept = self._state_accept_expecting_header
         self._in_buffer = ''
         self.chunk_in_progress = None
         self.chunks = collections.deque()
@@ -291,6 +294,8 @@
                 return 1
         elif self.state_accept == self._state_accept_reading_unused:
             return 1
+        elif self.state_accept == self._state_accept_expecting_header:
+            return max(0, len('chunked\n') - len(self._in_buffer))
         else:
             raise AssertionError("Impossible state: %r" % (self.state_accept,))
 
@@ -321,6 +326,19 @@
             self.error_in_progress = None
         self.finished_reading = True
 
+    def _state_accept_expecting_header(self, bytes):
+        self._in_buffer += bytes
+        prefix = self._extract_line()
+        if prefix is None:
+            # We haven't read a complete length prefix yet, so there's nothing
+            # to do.
+            return
+        elif prefix == 'chunked':
+            self.state_accept = self._state_accept_expecting_length
+        else:
+            raise errors.SmartProtocolError(
+                'Bad chunked body header: "%s"' % (prefix,))
+
     def _state_accept_expecting_length(self, bytes):
         self._in_buffer += bytes
         prefix = self._extract_line()

=== modified file 'bzrlib/tests/test_smart_transport.py'
--- bzrlib/tests/test_smart_transport.py	2007-11-09 19:21:56 +0000
+++ bzrlib/tests/test_smart_transport.py	2007-11-09 20:49:54 +0000
@@ -1681,13 +1681,14 @@
 
     def test_body_stream_serialisation_empty(self):
         """A body_stream with no bytes can be serialised."""
-        self.assertBodyStreamSerialisation('END\n', [])
+        self.assertBodyStreamSerialisation('chunked\nEND\n', [])
         self.assertBodyStreamRoundTrips([])
 
     def test_body_stream_serialisation(self):
         stream = ['chunk one', 'chunk two', 'chunk three']
         self.assertBodyStreamSerialisation(
-            '9\nchunk one' + '9\nchunk two' + 'b\nchunk three' + 'END\n',
+            'chunked\n' + '9\nchunk one' + '9\nchunk two' + 'b\nchunk three' +
+            'END\n',
             stream)
         self.assertBodyStreamRoundTrips(stream)
 
@@ -1698,7 +1699,7 @@
         """
         stream = ['', 'chunk']
         self.assertBodyStreamSerialisation(
-            '0\n' + '5\nchunk' + 'END\n', stream)
+            'chunked\n' + '0\n' + '5\nchunk' + 'END\n', stream)
         self.assertBodyStreamRoundTrips(stream)
 
     def test_body_stream_error_serialistion(self):
@@ -1706,7 +1707,7 @@
                   request.FailedSmartServerResponse(
                       ('FailureName', 'failure arg'))]
         expected_bytes = (
-            'b\nfirst chunk' +
+            'chunked\n' + 'b\nfirst chunk' +
             'ERR\n' + 'b\nFailureName' + 'b\nfailure arg' +
             'END\n')
         self.assertBodyStreamSerialisation(expected_bytes, stream)
@@ -1955,10 +1956,12 @@
             errors.ReadingCompleted, smart_protocol.read_body_bytes)
 
     def test_streamed_body_bytes(self):
+        body_header = 'chunked\n'
         two_body_chunks = "4\n1234" + "3\n567"
         body_terminator = "END\n"
         server_bytes = (protocol.RESPONSE_VERSION_TWO +
-                        "success\nok\n" + two_body_chunks + body_terminator)
+                        "success\nok\n" + body_header + two_body_chunks +
+                        body_terminator)
         input = StringIO(server_bytes)
         output = StringIO()
         client_medium = medium.SmartSimplePipesClientMedium(input, output)
@@ -1971,11 +1974,12 @@
 
     def test_read_streamed_body_error(self):
         """When a stream is interrupted by an error..."""
+        body_header = 'chunked\n'
         a_body_chunk = '4\naaaa'
         err_signal = 'ERR\n'
         err_chunks = 'a\nerror arg1' + '4\narg2'
         finish = 'END\n'
-        body = a_body_chunk + err_signal + err_chunks + finish
+        body = body_header + a_body_chunk + err_signal + err_chunks + finish
         server_bytes = (protocol.RESPONSE_VERSION_TWO +
                         "success\nok\n" + body)
         input = StringIO(server_bytes)
@@ -2097,18 +2101,23 @@
 
 
 class TestChunkedBodyDecoder(tests.TestCase):
-    """Tests for ChunkedBodyDecoder."""
+    """Tests for ChunkedBodyDecoder.
+    
+    This is the body decoder used for protocol version two.
+    """
 
     def test_construct(self):
         decoder = protocol.ChunkedBodyDecoder()
         self.assertFalse(decoder.finished_reading)
-        self.assertEqual(2, decoder.next_read_size())
+        self.assertEqual(8, decoder.next_read_size())
         self.assertEqual(None, decoder.read_next_chunk())
         self.assertEqual('', decoder.unused_data)
 
     def test_empty_content(self):
-        """'0' + LF is the complete chunked encoding of a zero-length body."""
+        """'chunked\nEND\n' is the complete encoding of a zero-length body.
+        """
         decoder = protocol.ChunkedBodyDecoder()
+        decoder.accept_bytes('chunked\n')
         decoder.accept_bytes('END\n')
         self.assertTrue(decoder.finished_reading)
         self.assertEqual(None, decoder.read_next_chunk())
@@ -2117,6 +2126,7 @@
     def test_one_chunk(self):
         """A body in a single chunk is decoded correctly."""
         decoder = protocol.ChunkedBodyDecoder()
+        decoder.accept_bytes('chunked\n')
         chunk_length = 'f\n'
         chunk_content = '123456789abcdef'
         finish = 'END\n'
@@ -2130,6 +2140,7 @@
         then we haven't finished reading yet.
         """
         decoder = protocol.ChunkedBodyDecoder()
+        decoder.accept_bytes('chunked\n')
         chunk_length = '8\n'
         three_bytes = '123'
         decoder.accept_bytes(chunk_length + three_bytes)
@@ -2145,6 +2156,7 @@
         """A chunk length hasn't been read until a newline byte has been read.
         """
         decoder = protocol.ChunkedBodyDecoder()
+        decoder.accept_bytes('chunked\n')
         decoder.accept_bytes('9')
         self.assertEqual(
             1, decoder.next_read_size(),
@@ -2161,6 +2173,7 @@
     def test_two_chunks(self):
         """Content from multiple chunks is concatenated."""
         decoder = protocol.ChunkedBodyDecoder()
+        decoder.accept_bytes('chunked\n')
         chunk_one = '3\naaa'
         chunk_two = '5\nbbbbb'
         finish = 'END\n'
@@ -2174,6 +2187,7 @@
     def test_excess_bytes(self):
         """Bytes after the chunked body are reported as unused bytes."""
         decoder = protocol.ChunkedBodyDecoder()
+        decoder.accept_bytes('chunked\n')
         chunked_body = "5\naaaaaEND\n"
         excess_bytes = "excess bytes"
         decoder.accept_bytes(chunked_body + excess_bytes)
@@ -2187,6 +2201,7 @@
     def test_multidigit_length(self):
         """Lengths in the chunk prefixes can have multiple digits."""
         decoder = protocol.ChunkedBodyDecoder()
+        decoder.accept_bytes('chunked\n')
         length = 0x123
         chunk_prefix = hex(length) + '\n'
         chunk_bytes = 'z' * length
@@ -2204,6 +2219,7 @@
         is called.
         """
         decoder = protocol.ChunkedBodyDecoder()
+        decoder.accept_bytes('chunked\n')
         chunk_length = 'f\n'
         chunk_content = '123456789abcdef'
         finish = 'END\n'
@@ -2216,6 +2232,7 @@
     def test_read_pending_data_resets(self):
         """read_pending_data does not return the same bytes twice."""
         decoder = protocol.ChunkedBodyDecoder()
+        decoder.accept_bytes('chunked\n')
         chunk_one = '3\naaa'
         chunk_two = '3\nbbb'
         finish = 'END\n'
@@ -2227,6 +2244,7 @@
 
     def test_decode_error(self):
         decoder = protocol.ChunkedBodyDecoder()
+        decoder.accept_bytes('chunked\n')
         chunk_one = 'b\nfirst chunk'
         error_signal = 'ERR\n'
         error_chunks = '5\npart1' + '5\npart2'
@@ -2238,6 +2256,14 @@
             ('part1', 'part2'))
         self.assertEqual(expected_failure, decoder.read_next_chunk())
 
+    def test_bad_header(self):
+        """accept_bytes raises a SmartProtocolError if a chunked body does not
+        start with the right header.
+        """
+        decoder = protocol.ChunkedBodyDecoder()
+        self.assertRaises(
+            errors.SmartProtocolError, decoder.accept_bytes, 'bad header\n')
+
 
 class TestSuccessfulSmartServerResponse(tests.TestCase):
 

=== modified file 'doc/developers/network-protocol.txt'
--- doc/developers/network-protocol.txt	2007-09-03 17:09:56 +0000
+++ doc/developers/network-protocol.txt	2007-11-09 20:49:54 +0000
@@ -184,9 +184,9 @@
 
 That is, the body consists of a series of chunks.  Each chunk starts with
 a length prefix in hexadecimal digits, followed by an ASCII newline byte.
-The end of the body is signaled by 'END\\n', or by 'ERR\\n' followed by
-error args, one per chunk.  Note that this allows an 8-bit clean error
-response.
+The end of the body is signaled by '``END\\n``', or by '``ERR\\n``'
+followed by error args, one per chunk.  Note that these args are 8-bit
+safe, unlike request args.
 
 A streamed body starts with the string "chunked" so that legacy clients
 and servers will not mistake the first chunk as the start of a version one

